from abc import ABC, abstractmethod

import numpy as np

from sklearn.base import BaseEstimator, ClassifierMixin
from sklearn.utils import check_random_state
from sklearn.utils.multiclass import check_classification_targets
from sklearn.utils.validation import check_is_fitted, check_array

from .. import distances
from .. import solvers
from ..solvers import SolverBaseClass
from .._utils import init_class

from ..distances import DistanceBaseClass
from ..objectives import GeneralizedLearningObjective

from typing import Tuple, Union, List

ModelParamsType = np.ndarray


class LVQBaseClass(
    ABC, BaseEstimator, ClassifierMixin
):  # lgtm [py/conflicting-attributes]
    """Learning Vector Quantization base class

    Abstract class for implementing LVQ models. It provides abstract methods with
    expected call signatures.

    Provides a common interface to the solver and other function that require access to the
    models. Additionally, it implements a number of functions shared by the currently implemented
    LVQ variations.

    See also
    --------
    GLVQ, GMLVQ, LGMLVQ
    """

    # Public attributes
    prototypes_: np.ndarray

    # "Private" attributes
    _distance: Union[DistanceBaseClass, object]
    _objective: GeneralizedLearningObjective
    _solver: Union[SolverBaseClass, object]

    # Related to model parameters
    _prototypes_size: int
    _prototypes_shape: Tuple

    # _variables stores the  model parameters (e.g., prototypes) in 1D format. The
    # get_prototypes() and other model parameter functions should return a view into this 1D
    # array.  Likewise set_prototypes() overwrites the _variables array.
    _variables: np.ndarray

    def __init__(
        self,
        distance_type: Union[str, type] = "squared-euclidean",
        distance_params: dict = None,
        valid_distances: List[str] = None,
        solver_type: Union[str, type] = "steepest-gradient-descent",
        solver_params: dict = None,
        valid_solvers: List[str] = None,
        prototype_init: Union[str, np.ndarray] = "class-conditional-mean",
        prototype_n_per_class: int = 1,
        random_state: Union[int, np.random.RandomState] = None,
        force_all_finite: Union[str, bool] = True,
    ):
        self.distance_type = distance_type
        self.distance_params = distance_params
        self.valid_distances = valid_distances

        self.solver_type = solver_type
        self.solver_params = solver_params
        self.valid_solvers = valid_solvers

        self.prototype_init = prototype_init
        self.prototype_n_per_class = prototype_n_per_class

        self.random_state = random_state
        self.force_all_finite = force_all_finite

    ###########################################################################################
    # The "Getter" and "Setter" that are used by the solvers to set and get model params.
    ###########################################################################################

    def get_variables(self) -> np.ndarray:
        r"""
        Returns the ``self._variables`` array that owns the memory allocated for the model
        parameters.

        Returns
        -------
        _variables : ndarray
            returns the model's _variables array.
        """
        return self._variables

    def set_variables(self, new_variables: np.ndarray) -> None:
        r"""
        Modifies the ``self._variables`` by copying the values of ``new_variables`` into the
        memory of ``self._variables``.

        Parameters
        ----------
        new_variables : ndarray
            1d numpy array that contains all the model parameters in continuous memory
        """
        np.copyto(self._variables, new_variables)

    @abstractmethod
    def get_model_params(self) -> Union[tuple, np.ndarray]:
        r"""
        Should return a view or tuple of views (in correct shape) of the model's parameters.
        Implementation depends on specific model as model parameters may differ per model.

        Returns
        -------
        ndarray or tuple
            View or tuple of views of the model's parameters.
        """
        raise NotImplementedError("You should implement this!")

    @abstractmethod
    def set_model_params(self, new_model_params: Union[tuple, np.ndarray]):
        r"""
        Should modify the ``self._variables`` array. Accepts the new_model_params in the shape of
        the model's parameters, e.g., prototypes or (prototypes, relevance_matrix).

        Always needs to be all the model parameters, can not be used for partial updates.

        Parameters
        ----------
        new_model_params : ndarray or tuple
            Array or tuple of arrays of the new model's parameters.
        """
        raise NotImplementedError("You should implement this!")

    ###########################################################################################
    # Specific "getters" and "setters" for the prototypes shared by every LVQ model.
    ###########################################################################################

    def get_prototypes(self) -> np.ndarray:
        r"""
        Return a view into ``self._variables`` of the the shape of the prototypes (n_prototypes,
        n_features). At the moment only consistency function, does not actually  create the shape
        and only  works after ``self.prototypes_`` has been set.

        Returns
        -------
        prototypes : ndarray of shape (n_prototypes, n_features)
            View into ``self._variables`` with shape specified above.
        """
        return self.prototypes_

    def set_prototypes(self, new_prototypes: np.ndarray) -> None:
        r"""
        Accepts a new_prototypes array with the same shape as ``self.prototypes_`` and overwrites
        the ``self._variables`` array by copying the values of the new_prototypes.

        Parameters
        ----------
        new_prototypes : ndarray of shape (n_prototypes, n_features)
            The new prototypes the model should store.
        """
        np.copyto(self.prototypes_, new_prototypes)

    ###########################################################################################
    # Functions to transform the 1D variables array to model parameters and back
    ###########################################################################################

    @abstractmethod
    def to_model_params_view(self, var_buffer: np.ndarray) -> Union[tuple, np.ndarray]:
        r"""
        Should return a single view into the var_buffer or a tuple of views. This depends on the
        model and its parameters.

        Parameters
        ----------
        var_buffer : ndarray
            Array with the same size as the model's variables array as returned
            by ``get_variables()``.

        Returns
        -------
        ndarray or tuple
            Should return a view or tuple of views of the model parameters in appropriate shapes.
        """
        raise NotImplementedError("You should implement this!")

    @abstractmethod
    def to_prototypes_view(self, var_buffer: np.ndarray) -> np.ndarray:
        r"""
        Should return the prototypes from the provided var_buffer. I.e., it selects/views the
        appropriate part of memory and reshapes it.

        Parameters
        ----------
        var_buffer : ndarray
            Array with the same size as the model's variables array as returned
            by ``get_variables()``.

        Returns
        -------
        ndarray of shape (n_prototypes, n_features)
            View into the var_buffer.
        """
        raise NotImplementedError("You should implement this!")

    ###########################################################################################
    # Solver Normalization functions
    ###########################################################################################

    @abstractmethod
    def normalize_variables(self, var_buffer: np.ndarray) -> None:
        r"""
        Should modify the var_buffer as if it was the variables array provided
        by ``get_variables()``.

        Parameters
        ----------
        var_buffer : ndarray
            Array with the same size as the model's variables array as returned
            by ``get_variables()``.


        Returns
        -------
        ndarray or tuple
            Same shape and size as input, but normalized. How to normalize depends on model
            implementation.
        """
        raise NotImplementedError("You should implement this!")

    @staticmethod
    def _normalize_prototypes(prototypes: np.ndarray) -> None:
        r"""
        Normalizes the provided prototypes array, i.e., it writes to the same memory. Performs
        the following normalization step for each prototype :math:`\mathbf{w_i}`:

        ..math::
             \mathbf{w}_i / || \mathbf{w}_i ||

        Parameters
        ----------
        prototypes : ndarray of shape (n_prototypes, n_features)
            To be normalized prototypes.

        Returns
        -------
        ndarray of same shape as input
            Normalized prototypes.
        """
        np.divide(
            prototypes,
            np.linalg.norm(prototypes, axis=1, keepdims=True),
            out=prototypes,
        )

    ###########################################################################################
    # Solver helper functions
    ###########################################################################################

    @abstractmethod
    def add_partial_gradient(self, gradient, partial_gradient, i_prototype) -> None:
        r"""
        To increase performance, the distance gradient methods (should) return only the relevant values.
        I.e., the gradient of the prototype i_prototype and potentially other parameters linked to
        this prototype. This partial gradient needs to added (overwrite) to the correct parts of
        the actual gradient and this is what this function should do.

        Parameters
        ----------
        gradient : ndarray
            Same shape as the ``get_variables()`` would return.

        partial_gradient : ndarray
            1d array containing the partial gradient.

        i_prototype : int
            The index of the prototype to which the partial gradient was  computed.
        """
        raise NotImplementedError("You should implement this!")

    @abstractmethod
    def mul_step_size(
        self, step_size: Union[float, np.ndarray], gradient: np.ndarray
    ) -> None:
        r"""
        Should multiply the provided gradient with the provided step size and overwrite the
        values in ``gradient``.  Depending on the ``step_size`` being a float or array different
        step sizes are used for different model parameters (which also depends on the model if
        there are more then only prototypes)

        Parameters
        ----------
        step_size : float or ndarray
            The scalar or list of values containing the step sizes.
        gradient : ndarray
            Same shape as the ``get_variables()`` would return.
        """
        raise NotImplementedError("You should implement this!")

    ###########################################################################################
    # Initialization function
    ###########################################################################################

    @abstractmethod
    def _init_variables(self):
        r"""
        Should initialize the variables, 1d numpy array to hold all model parameters. Should
        store these in self._variables.
        """
        raise NotImplementedError("You should implement this!")

    @abstractmethod
    def _check_model_params(self):
        r"""
        Should check the model parameters. I.e., call check_prototype_params with parameters and
        other model parameters that there might be.
        """

    @abstractmethod
    def _init_model_params(self, X, y):
        r"""
        Depending on the model, things such as self.prototypes_ should be initialized and set
        using the set methods or one should make sure the parameters are views into the variables
        array, such that variables array is changed as well.

        Parameters
        ----------
        X : ndarray, with shape (n_samples, n_features)
            The X
        y : ndarray, with shape (n_samples)
            The labels
        """
        raise NotImplementedError("You should implement this!")

    def _check_prototype_params(self):
        """
        Check prototype params, i.e., if the prototypes_per_class is set correctly.
        Additionally, it sets the size and shape of the prototypes such that these can be used
        for the creation of the ``self._variables`` and view ``self.prototypes_``.
        """
        prototype_n_per_class = self.prototype_n_per_class

        if isinstance(prototype_n_per_class, int):
            self._prototypes_shape = (
                prototype_n_per_class * self.classes_.size,
                self.n_features_in_,
            )
        elif isinstance(prototype_n_per_class, np.ndarray):
            if prototype_n_per_class.size != self.classes_.size:
                raise ValueError(
                    "Expected the number protoypes_per_class (size = {}) to have a number of elements "
                    "equal to the number of classes (size = {}).",
                    prototype_n_per_class.size,
                    self.classes_.size,
                )

            if np.any(prototype_n_per_class <= 0.0):
                raise ValueError(
                    "Prototypes_per_class ({}) cannot contain any values less than or equal to zero.",
                    prototype_n_per_class,
                )

            self._prototypes_shape = (
                np.sum(prototype_n_per_class),
                self.n_features_in_,
            )
        else:
            raise ValueError(
                "Expected prototypes_per_class to be either of type int or np.ndarray, but got type: {}",
                type(prototype_n_per_class),
            )

        self._prototypes_size = np.prod(self._prototypes_shape)

    def _init_prototypes(
        self,
        X: np.ndarray,
        y: np.ndarray,
    ) -> None:
        """
        Initialized the prototypes, with a small random offset, to the class conditional mean.
        To be used in the _initialize_parameters function.

        Parameters
        ----------
        X : ndarray with shape (number of observations, number of dimensions)
        y : ndarray with size equal to the number of observations
        """
        self.prototypes_labels_ = np.repeat(
            np.arange(0, self.classes_.size), self.prototype_n_per_class
        )

        # Sets initial value for prototypes....
        self.prototypes_ = self.to_prototypes_view(self._variables)

        if self.prototype_init == "class-conditional-mean":
            conditional_mean = _conditional_mean(self.prototypes_labels_, X, y)

            self.set_prototypes(
                conditional_mean
                + (1e-4 * self.random_state_.uniform(-1, 1, conditional_mean.shape))
            )
        else:
            raise ValueError(
                "The provided value for the parameter 'prototype_init' is invalid."
            )

    @abstractmethod
    def _init_objective(self) -> None:
        """
        Should initialize the ``self._objective``. Depends on the algorithm.
        """
        raise NotImplementedError("You should implement this!")

    def _init_distance(self) -> None:
        """
        Initializes the ``self.distance``.
        """
        distance_class = init_class(
            distances,
            self.distance_type,
            valid_class_types=self.valid_distances,
        )
        if self.distance_params is not None:
            self._distance = distance_class(**self.distance_params)
        else:
            self._distance = distance_class()

    def _init_solver(self) -> None:
        """
        Should initialize the ``self._solver``. Depends on the algorithm.
        """
        solver_class = init_class(solvers, self.solver_type, self.valid_solvers)
        if self.solver_params is not None:
            self._solver = solver_class(self._objective, **self.solver_params)
        else:
            self._solver = solver_class(self._objective)

    ###########################################################################################
    # Data and label validation
    ###########################################################################################

    def _check_data_and_labels(
        self, X: np.ndarray, labels: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        Functions performs a series of check mostly by using sklearn util functions.
        Additionally, it transform the labels into indexes of unique labels, which are stored in
        self.classes_.

        Checks X and labels for consistent length, enforces X to be 2D and labels
        1D. By default, X is checked to be non-empty and containing only finite values. Standard
        input checks are also applied to labels, such as checking that labels does not have
        np.nan or np.inf targets.
            - sklearn check_X_y()

        Ensure that target labels are of a non-regression type.
        Only the following target types (as defined in sklearn.utils.multiclass.type_of_target)
        are allowed:
            'binary', 'multiclass', 'multiclass-multioutput',
            'multilabel-indicator', 'multilabel-sequences'.
            - sklearn check_classification_targets()

        Parameters
        ----------
        X : ndarray of shape (number of observations, number of dimensions)
        labels : ndarray of size (number of observations)

        Returns
        -------
        X : ndarray with same shape (and values) as input
        labels : ndarray of indexes to self.classes_
        """
        # Check X
        X, labels = self._validate_data(
            X, labels, force_all_finite=self.force_all_finite
        )

        # Check classification targets
        check_classification_targets(labels)

        # Store unique provided labels in self.classes_ and transform the labels into a index
        # that can be used to reconstruct the original labels.
        self.classes_, labels = np.unique(labels, return_inverse=True)

        # Raise an error when the targets only contain a single unique class.
        if self.classes_.size <= 1:
            raise ValueError("Classifier can't train when only one class is present.")

        return X, labels

    ###########################################################################################
    # Before and after fit/solve function (initialization of things)
    ###########################################################################################

    def _before_fit(self, X: np.ndarray, y: np.ndarray):
        """
        Should initialize:
            1. self._variables and algorithm specific parameters which should be
               views into self._variables (share memory).
            2. The distance function in self.distance
            3. The objective function in self._objective
            4. The solver function in self._solver

        Parameters
        ----------
        X : ndarray, with shape (n_samples, n_features)
            The X
        y : ndarray, with shape (n_samples)
            The labels
        """
        self._check_model_params()

        # Initializes the 1D block of continuous memory to hold the model params.
        self._init_variables()

        # Initialize algorithm specific parameters.
        self._init_model_params(X, y)

        self._init_distance()

        self._init_objective()

        self._init_solver()

    def _after_fit(self, X: np.ndarray, y: np.ndarray):
        """
        Method that by default does nothing but can be used by methods that need to compute
        transformation matrices, such that this does not need to be checked or done everytime
        transform() is called (e.g., for GMLVQ and LGMLVQ)

        Parameters
        ----------
        X : ndarray with shape (number of observations, number of dimensions)
        y : ndarray with size equal to the number of observations
        """

    ###########################################################################################
    # Public API functions
    ###########################################################################################

    def fit(self, X: np.ndarray, y: np.ndarray):
        """Fit function that provides the general implementation of the LVQ algorithms. It checks the data, calls
         before_fit method, calls the solve method of the solver, and the after_fit method.

        Parameters
        ----------
        X : ndarray of shape (number of observations, number of dimensions)
        y : ndarray of size (number of observations)

        Returns
        -------
        self
            The trained model
        """
        # Check X and check and transform labels.
        X, y_index = self._check_data_and_labels(X, y)

        # Initialize random_state_ that should be used to perform any rng.
        self.random_state_ = check_random_state(self.random_state)

        # Before solve (handles initialization of things)
        self._before_fit(X, y_index)

        self._solver.solve(X, y_index, self)

        # After solve (handles initialization of things that can only be done after fit)
        self._after_fit(X, y_index)

        return self

    def _multiclass_decision_function(self, X: np.ndarray):
        """
        Computes the decision values and returns shape (n_observations, n_classes). The values
        are constructed by computing: the distance between an observation and the prototype with
        a different label minus the distance between that observation and the closest prototype
        with the same label, where "the same" is defined to be the index of the column the value
        is being computed for.

        Parameters
        ----------
        X : ndarray
            The data.

        Returns
        -------
        ndarray of shape (n_observations, n_classes)
        """
        # Of shape n_observations , n_prototypes
        distances = self._distance(X, self)

        # Allocation n_observations, n_classes
        decision_values = np.zeros((X.shape[0], self.classes_.size))

        # return n_observations, n_classes
        for i, _ in enumerate(self.classes_):
            decision_values[:, i] = distances[:, self.prototypes_labels_ != i].min(
                axis=1
            ) - distances[:, self.prototypes_labels_ == i].min(axis=1)

        return decision_values

    def decision_function(self, X: np.ndarray):
        """
        Evaluates the decision function for the samples in X.
        Shape for binary class is (n_observations,) with the decision values for the "greater"
        class. In the multiclass case it returns decision values for each class and therefore has
        the shape (n_observations, n_classes).

        Parameters
        ----------
        X : ndarray
            The data.

        Returns
        -------
        decision_values : ndarray
            Binary case shape is (n_observations,) and the multiclass case (n_observations,
            n_classes)
        """
        # SciKit-learn list of checked params before predict
        check_is_fitted(self)

        # Input validation
        X = check_array(X, force_all_finite=self.force_all_finite)

        decision_values = self._multiclass_decision_function(X)

        if self.classes_.size == 2:
            # The decision function needs to return (n_samples,) for the "greater" class.
            return decision_values[:, 1]

        # else it should  return (n_samples, n_classes)
        return decision_values

    def predict_proba(self, X: np.ndarray):
        """
        Parameters
        ----------
        X  : ndarray
            The data.

        Returns
        -------
        confidence_scores : ndarray of shape (n_observations, n_classes)
        """
        # SciKit-learn list of checked params before predict
        check_is_fitted(self)

        # Input validation
        X = check_array(X, force_all_finite=self.force_all_finite)

        # Between -1  and 1
        decision_values = self._multiclass_decision_function(X)

        # Softmax function (keeps the same scipy.stats.rankdata)
        # Very  arbitrary  0.01, which also might not always work?
        exp_decision_values = np.exp(0.01 * decision_values)

        return exp_decision_values / np.sum(exp_decision_values, axis=1)[:, np.newaxis]

    def predict(self, X: np.ndarray):
        """Predict function

        The decision is made for the label of the prototype with the minimum decision value,
        as provided by the ``decision_function()``.

        Parameters
        ----------
         X  : ndarray
            The data.

        Returns
        -------
        ndarray of shape (n_observations)
            Returns the predicted labels.
        """
        # SciKit-learn list of checked params before predict
        check_is_fitted(self)

        # Input validation
        # X = self._validate_data(X, force_all_finite=self.force_all_finite)
        decision_values = self.decision_function(X)

        if self.classes_.size == 2:
            return self.classes_[(decision_values > 0).astype(np.int)]

        # Lower value is the closest prototype.
        return self.classes_[decision_values.argmax(axis=1)]


def _conditional_mean(p_labels: np.ndarray, data: np.ndarray, d_labels: np.ndarray):
    """ Implements the conditional mean (ignoring nan values), i.e., mean per class"""
    return np.array(
        [np.nanmean(data[p_label == d_labels, :], axis=0) for p_label in p_labels]
    )



from typing import TYPE_CHECKING, Union

import numpy as np
from sklearn.utils import shuffle

from . import SolverBaseClass
from ._base import _update_state
from ..objectives import ObjectiveBaseClass

if TYPE_CHECKING:
    from sklvq.models import LVQBaseClass

STATE_KEYS = ["variables", "nit", "fun", "step_size"]


class SteepestGradientDescent(SolverBaseClass):
    r"""Steepest gradient descent (SGD)

    Implements the steepest gradient descent optimization method. Can perform stochastic,
    mini-batch and batch gradient descent by changing the batch_size. Implementation is
    inspired by the description given in `[1]`_.

    The algorithm performs the following update of the model parameters (:math:`\mathbf{\theta}`) per
    batch. This process is repeated multiple times (per step) when the ``batch_size`` (:math:`M`)
    is smaller than the total number of samples in the data.

    .. math::
        \mathbf{\theta} = \mathbf{\theta}  - \eta(t) \cdot \sum_i^M \nabla  e_i(\mathbf{\theta}),

    with :math:`\nabla e_i(\mathbf{\theta})` the gradient of the objective function with respect to
    a sample given the current model parameters :math:`\mathbf{\theta}`, and :math:`\eta(t)` the step
    size at step :math:`t`, which is changed using a simple annealing function:

    .. math::
        \eta(t) = \frac{\eta_{init}} {(1 + \frac{t}{t_{max}})},

    with :math:`t_{max}` given by the ``max_runs`` parameter and :math:`\eta_{init}` by the
    ``step_size`` parameter.


    Parameters
    ----------
    objective: ObjectiveBaseClass, required
        This is set by the algorithm. See :class:`sklvq.models.GLVQ`, :class:`sklvq.models.GMLVQ`,
        and :class:`sklvq.models.LGMLVQ`.

    max_runs: int
        Maximum number of runs/epochs that will be computed. Should be >= 1. Early stopping can
        be implemented by providing a ``callback`` function that returns True when the solver should
        stop.

    batch_size: int
        Controls the batch size and accepts a value >= 0. The value indicates the number of
        samples considered to be in the batch. A stochastic gradient descent corresponds with a
        ``batch_size`` of 1. For Batch gradient descent 0 can be used to indicate to use all the
        samples. Any value > 1 < n_samples can be considered as a mini-batch gradient descent.

        If batches can not properly  be divided in batches with the specified size the last
        batch might contain less than the specified number of samples.

        The data is always shuffled before it is split into batches.

    step_size: float or ndarray
        The step size to control the learning rate of the model parameters. If the same step size
        should be used for all parameters (e.g., prototypes and omega) then a single float is
        sufficient. If separate initial step sizes should be used per model parameter then this
        should be specified by using a numpy array.

    callback: callable
        Callable with signature callable(state). If the callable returns True the solver
        will stop even if ``max_runs`` is not reached yet. The state object contains the following:

        - "variables"
            Concatenated 1D ndarray of the model's parameters
        - "nit"
            The current iteration counter
        - "fun"
            The objective cost
        - "step_size"
            The current step_size(s)

    References
    ----------
    _`[1]` LeKander, M., Biehl, M., & De Vries, H. (2017). "Empirical evaluation of gradient
    methods for matrix learning vector quantization." 12th International Workshop on
    Self-Organizing Maps and Learning Vector Quantization, Clustering and Data
    Visualization, WSOM 2017.
    """

    def __init__(
        self,
        objective: ObjectiveBaseClass,
        max_runs: int = 10,
        batch_size: int = 1,
        step_size: Union[float, np.ndarray] = 0.1,
        callback: callable = None,
    ):
        super().__init__(objective)
        if max_runs <= 0:
            raise ValueError(
                "{}:  Expected max_runs to be > 0, but got max_runs = {}".format(
                    type(self).__name__, max_runs
                )
            )
        self.max_runs = max_runs

        if batch_size < 0:
            raise ValueError(
                "{}:  Expected batch_size to be >= 0, but got batch_size = {}".format(
                    type(self).__name__, batch_size
                )
            )
        self.batch_size = batch_size

        if np.any(step_size <= 0):
            raise ValueError(
                "{}:  Expected step_size to be > 0, but got step_size = {}".format(
                    type(self).__name__, step_size
                )
            )
        self.step_size = step_size

        if callback is not None:
            if not callable(callback):
                raise ValueError(
                    "{}:  callback is not callable.".format(type(self).__name__)
                )
        self.callback = callback

    def solve(
        self,
        data: np.ndarray,
        labels: np.ndarray,
        model: "LVQBaseClass",
    ):
        """Solve function that gets called by the fit method of the models.

        Performs the steps of the steepest gradient descent optimization method.

        Parameters
        ----------
        data : ndarray of shape (n_samples, n_features)
            The data.

        labels : ndarray of size (n_samples)
            The labels of the samples in the data.

        model : LVQBaseClass
            The initial model that will also hold the final result
        """

        if self.callback is not None:
            state = _update_state(
                STATE_KEYS,
                variables=np.copy(model.get_variables()),
                nit="Initial",
                fun=self.objective(model, data, labels),
            )
            if self.callback(state):
                return

        batch_size = self.batch_size

        # Less than 0 is caught in init.
        if batch_size > data.shape[0]:
            raise ValueError("Provided batch_size is invalid.")
        elif batch_size == 0:
            batch_size = data.shape[0]

        for i_run in range(0, self.max_runs):
            # Randomize order of samples
            shuffled_indices = shuffle(
                np.array(range(0, labels.size)), random_state=model.random_state_
            )

            # Divide the shuffled indices into batches (not necessarily equal size,
            # see documentation of numpy.array_split).
            batches = np.array_split(
                shuffled_indices,
                list(range(batch_size, labels.size, batch_size)),
                axis=0,
            )

            # Update step size using a simple annealing strategy
            step_size = self.step_size / (1 + i_run / self.max_runs)

            for i_batch in batches:
                # Creating views when batch_size is 1 does not do much for speed.
                batch = data[i_batch, :]
                batch_labels = labels[i_batch]

                # Compute objective gradient
                objective_gradient = self.objective.gradient(model, batch, batch_labels)

                # Multiply each param by its given step_size
                model.mul_step_size(step_size, objective_gradient)

                # Update the model by subtracting the objective-gradient (descent) from the
                # current models variables, e.g., (prototypes, omega) in case of GMLVQ
                model.set_variables(
                    np.subtract(  # returns out=objective_gradient
                        model.get_variables(),
                        objective_gradient,
                        out=objective_gradient,
                    )
                )

            if self.callback is not None:
                state = _update_state(
                    STATE_KEYS,
                    variables=np.copy(model.get_variables()),
                    nit=i_run + 1,
                    fun=self.objective(model, data, labels),
                    step_size=step_size,
                )
                if self.callback(state):
                    return



from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

import numpy as np

if TYPE_CHECKING:
    from ..models import LVQBaseClass


class DistanceBaseClass(ABC):
    """Distance base class

    Abstract class for implementing distance functions. It provides abstract methods with
    expected call signatures.

    Custom distance function '__init__' should accept any parameters as key-value pairs.

    See also
    --------
    Euclidean, SquaredEuclidean, AdaptiveSquaredEuclidean, LocalAdaptiveSquaredEuclidean
    """

    __slots__ = ()

    @abstractmethod
    def __call__(self, data: np.ndarray, model: "LVQBaseClass") -> np.ndarray:
        """
        Should implement a distance function.

        Parameters
        ----------
        data : ndarray with shape (n_samples, n_features)
            The samples for which the distance to the prototypes of the model need to be computed.
        model : LVQBaseClass
            Any class extending the LVQBaseClass or depending on the type of distance function
            is implemented a class that provides the required attributes.

        Returns
        -------
        ndarray with shape (n_samples, n_prototypes)
            Evaluation of the distance between each sample and prototype of the model.
        """
        raise NotImplementedError("You should implement this!")

    @abstractmethod
    def gradient(
        self, data: np.ndarray, model: "LVQBaseClass", i_prototype: int
    ) -> np.ndarray:
        """The distance gradient method.

        Parameters
        ----------
        data : ndarray with shape (n_samples, n_features)
            The data for which the distance gradient to the prototypes of the model need to be
            computed.
        model : LVQBaseClass
            Any class extending the LVQBaseClass or depending on the type of distance function
            is implemented, a class that provides the required attributes.
        i_prototype : int
            The index of the prototype for which the gradient needs to be computed.

        Returns
        -------
        ndarray with shape (n_samples, n_features)
            The gradient with respect to the prototype (i_prototype) and every sample in X.

        """
        raise NotImplementedError("You should implement this!")



import numpy as np

from .test_common import check_distance
from .test_common import check_init_distance

from sklvq.models import GLVQ


def test_squared_euclidean():
    check_init_distance("squared-euclidean")

    data = np.array([[1, 2, 3], [-1, -2, -3], [0, 0, 0]], dtype="float64")
    p = np.array([[1, 2, 3], [-1, -2, -3], [0, 0, 0]])

    model = GLVQ(distance_type="squared-euclidean")
    model.fit(data, np.array([0, 1, 2]))
    model.set_prototypes(p)

    check_distance(model._distance, data, model)

    # Check force_all_finite settings

    model = GLVQ(distance_type="squared-euclidean", force_all_finite="allow-nan")
    model.fit(data, np.array([0, 1, 2]))
    model.set_prototypes(p)

    data[0, 0] = np.nan
    data[1, 0] = np.nan

    check_distance(model._distance, data, model)



import numpy as np

from .test_common import check_init_activation

import pytest


def test_sigmoid():
    sigmoid_class = check_init_activation("sigmoid")

    sigmoid = sigmoid_class(beta=2)

    assert sigmoid.beta == 2

    sigmoid = sigmoid_class()

    assert sigmoid.beta == 1

    wrong_betas = [-1, 0]
    for wrong_beta in wrong_betas:
        with pytest.raises(ValueError):
            sigmoid_class(beta=wrong_beta)

    # Check if x = 0 equals approximately 0.5
    assert sigmoid(np.array([0])) == pytest.approx(0.5)
    # Check if at x = 0 the gradient equals approximately 0.25
    assert sigmoid.gradient(np.array([0])) == pytest.approx(0.25)

    # "Always" positive with negative numbers
    assert np.all(sigmoid(np.array([-1, -10, -100, -1000, -10000]) > 0))
    # "Always" positive with positive numbers
    assert np.all(sigmoid(np.array([1, 10, 100, 1000, 10000]) > 0))

    # Symmetry
    assert np.all(
        pytest.approx(sigmoid(np.array([1, 2, 3, 4])))
        == (1 - sigmoid(np.array([-1, -2, -3, -4])))
    )

    sigmoid = sigmoid_class(beta=10)

    # Check if x = 0 is not changed
    assert sigmoid(np.array([0])) == pytest.approx(0.5)
    # Check if x = 0 is changed based on beta.
    assert sigmoid.gradient(np.array([0])) == pytest.approx(2.5)

    # Potentially raising errors when division by 0.0 or inf happens....
